require "reactphysics3d"

-- note: we will use 3D support library in Library/3dbase folder of gideros installation

-- create physics world
local world=r3d.World.new(0,-9.8,0)

-- set up a fullscreen 3d viewport
local sw,sh=application:getContentWidth(),application:getContentHeight()
local view=D3.View.new(sw,sh,45,0.1,1000) -- 45 is the Field Of View, 0.1 is the near plane and 1000 the far plane

-- build a generic cube mesh with normals
-- unit cube, extending from -1 to +1 in each direction
local GENCUBE=D3.Cube.new() 
GENCUBE:mapTexture(Texture.new("box.png",true))
GENCUBE:updateMode(D3.Mesh.MODE_LIGHTING|D3.Mesh.MODE_SHADOW)
GENCUBE.shape=r3d.BoxShape.new(1,1,1)

-- build a 100x100 floor plane (with normals)
local gplane=D3.Mesh.new()
gplane:setVertexArray{-100,1,-100, 100,1,-100, 100,1,100, -100,1,100}
local tw,th=3200,3200 
gplane:setTextureCoordinateArray{0,0,tw,0,tw,th,0,th}
gplane:setGenericArray(3,Shader.DFLOAT,3,4,{
	0,1,0,0,1,0,0,1,0,0,1,0,
})
gplane:setIndexArray{1,2,3,1,3,4}
gplane:setTexture(Texture.new("grass.png",true,{wrap=TextureBase.REPEAT}))
gplane:updateMode(D3.Mesh.MODE_LIGHTING|D3.Mesh.MODE_SHADOW|D3.Mesh.MODE_TEXTURE)
-- build a reactphysics3d body for the plane
gplane.body=world:createBody()
gplane.body:setType(r3d.Body.STATIC_BODY)
gplane.body:createFixture(r3d.BoxShape.new(100,1,100),nil,1000)

-- function to build a cube instance
world.bodies={}
function build_cube(size)
	-- we actually build a viewport onto the generic cube we set up earlier
	local v=Viewport.new()
	v:setContent(GENCUBE)
	-- and randomly place/rotate it
	v:setPosition(math.random()*9-5,math.random()*10+10,math.random()*9-5)
	v:setRotation(math.random()*20-10)
	v:setRotationX(math.random()*20-10)
	v:setRotationY(math.random()*20-10)
	-- create body accordingly
	local body=world:createBody(v:getMatrix())
	body:createFixture(GENCUBE.shape,nil,1)
	v.body=body
	-- add the body to our moving bodies list
	world.bodies[v]=v.body
	return v
end

-- a player1 we can move
local player1=Viewport.new()
player1:setContent(GENCUBE)
player1:setColorTransform(0, 0, 1, 0.5)
player1:setPosition(0, 3, -5)
player1.body=world:createBody(player1:getMatrix())
player1.body:setAngularLockAxisFactor(0, 0, 0) -- 0=locked on axis
player1.body:setLinearDamping(1) -- 0 = no damping
local fixture = player1.body:createFixture(GENCUBE.shape,nil,2) -- shape,transform,mass
-- here we can modifiy the player1 materials
local mat = fixture:getMaterial()
mat.bounciness = 0.5 -- 0 = no bounciness, 1 = max bounciness
mat.frictionCoefficient = 2 -- 0 = no resistance
fixture:setMaterial(mat)
-- keyboard
stage:addEventListener(Event.KEY_DOWN, function(e)
	if e.keyCode == KeyCode.LEFT then player1.isleft = true end
	if e.keyCode == KeyCode.RIGHT then player1.isright = true end
	if e.keyCode == KeyCode.UP then player1.isup = true end
	if e.keyCode == KeyCode.DOWN then player1.isdown = true end
end)
stage:addEventListener(Event.KEY_UP, function(e)
	if e.keyCode == KeyCode.LEFT then player1.isleft = false end
	if e.keyCode == KeyCode.RIGHT then player1.isright = false end
	if e.keyCode == KeyCode.UP then player1.isup = false end
	if e.keyCode == KeyCode.DOWN then player1.isdown = false end
end)

-- setup our scene: add our plane
local scene=view:getScene()
scene:addChild(gplane)
scene:addChild(player1)

-- debug draw
--local debugDraw = r3d.DebugDraw.new(world)
--scene:addChild(debugDraw)

-- configure light
Lighting.setLight(15,30,0,0.3)
Lighting.setLightTarget(0,0,0,40,120)

-- setup camera
view:lookAt(0,10,-20,0,5,0)

-- setup the stage
-- add a blue gradient background for the sky
local sky=Pixel.new(0xFFFFFF,1,320*3,480*3)
sky:setColor(0x00FFFF,1,0x0040FF,1,90) sky:setPosition(-320,-240-480)
stage:addChild(sky)

-- add our 3d view on top of that
stage:addChild(view)

-- main loop:
local speed = 16
local gen=0
stage:addEventListener(Event.ENTER_FRAME,function(e)
	Lighting.computeShadows(scene)
	gen=gen+1
	-- create new box every x milliseconds
	if gen==200 then
		scene:addChild(build_cube(1))	
		gen=0
	end
	world:step(e.deltaTime)
	-- move the player1
	if player1.isleft and not player1.isright then
		player1.body:applyLocalForceAtCenterOfMass(speed, 0, 0)
	end
	if player1.isright and not player1.isleft then
		player1.body:applyLocalForceAtCenterOfMass(-speed, 0, 0)
	end
	if player1.isup and not player1.isdown then
		player1.body:applyLocalForceAtCenterOfMass(0, 0, speed)
	end
	if player1.isdown and not player1.isup then
		player1.body:applyLocalForceAtCenterOfMass(0, 0, -speed)
	end
	-- align view to body
	player1:setMatrix(player1.body:getTransform())
	for s,b in pairs(world.bodies) do -- view, body
		s:setMatrix(b:getTransform())
	end
	-- collisions
	for s,b in pairs(world.bodies) do
		world:testCollision(b, player1.body, function()
			world.bodies[s] = nil -- remove from list
			world:destroyBody(b) -- remove from the collision world
			scene:removeChild(s) -- remove from the scene
		end)
	end
end)

-- infos
local tf = TextField.new(nil, "use the arrow keys to move\nthe player")
tf:setTextColor(vector(1, 1, 0, 1))
tf:setScale(2)
tf:setPosition(8, 32)
stage:addChild(tf)
